import javax.sound.sampled.*;
import javax.swing.*;
import javax.swing.Timer;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

/**
 * <h1>PhoneKeypadExperiment</h1>
 * 
 * <h3>Summary</h3>
 * 
 * <ul>
 * <li>Experiment software to implement a text entry experiment using a
 * phone-style soft keyboard.
 * <p>
 * 
 * <li>Four text entry modes are supported:
 * <ul>
 * <li>Multitap (MT)
 * <li>T9 (T9)
 * <li>T9 with word completion (T9WC)
 * <li>T9 with word completion and word prediction (T9WCWP)
 * </ul>
 * <p>
 * </ul>
 * 
 * <h3>Related References</h3>
 * 
 * <ul>
 * <li>
 * <a href="http://www.yorku.ca/mack/chi00.html"> Predicting text entry speed on mobile phones </a>,
 * by Silfverberg, MacKenzie, and Korhonen (<i>CHI '99</i>)
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/hcimobile02.html"> KSPC (keystrokes per character) as a
 * characteristic of text entry techniques </a>, MacKenzie (<i>MobileHCI 2002</i>)
 * <p>
 * 
 * <li>
 * <a href="http://www.yorku.ca/mack/chi03b.html">Phrase sets for evaluating text entry
 * techniques</a>, MacKenzie and Soukoreff (<i>CHI 2003</i>)
 * </ul>
 * <p>
 * 
 * <h3>Running the Experiment Software</h3>
 * <p>
 * 
 * <a href="http://www.yorku.ca/mack/HCIbook/Running/">Click here</a> for instructions on
 * launching/running the application.
 * <p>
 * 
 * <h3>Setup Parameters</h3>
 * 
 * Upon launching, the program presents a setup dialog:
 * <p>
 * 
 * <center><a href="PhoneKeypadExperiment-0.jpg"><img src="PhoneKeypadExperiment-0.jpg"
 * width="400"></a></center>
 * <p>
 * 
 * The default parameter settings are read from a configuration file called
 * <code>PhoneKeypadExperiment.cfg</code>. This file is created automatically when the application
 * is launched for the first time. The default parameter settings may be changed through the setup
 * dialog. The setup parameters are as follows:
 * <p>
 * 
 * <blockquote>
 * <table border="1" cellspacing="0" cellpadding="6" valign="top">
 * <tr bgcolor="#cccccc">
 * <th>Parameter
 * <th>Description
 * 
 * <tr valign="top">
 * <td>Participant Code
 * <td>Identifies the current participant. This is used in forming the names for
 * the output data files. Also, the sd2 output data file includes a column with the participant
 * code.
 * <p>
 * 
 * <tr valign="top">
 * <td>Condition Code
 * <td>An arbitrary code used to associate a test condition with this invocation.
 * This parameter might be useful if the software is used in an experiment, since conditions might
 * be tested that are not inherently part of the application (e.g., Gender &rarr; male, female). The
 * condition code is used in forming the name for the output data file. Also, the sd2 output data
 * file contains a column with the condition code.
 * <p>
 * 
 * <tr valign="top">
 * <td>Session Code
 * <td>Identifies the session. Useful if testing proceeds over multiple sessions to
 * gauge the progression of learning. The session code is used in forming the name
 * for the output data file. Also, the sd2 output data file contains a column with the condition
 * code.
 * <p>
 * 
 * Note: There is no setup parameter for "Block code". The block code is generated automatically by the
 * software.
 * <p>
 *  
 * <tr valign="top">
 * <td>Number of Phrases
 * <td>Specifies the number of phrases presented to the participant in the
 * current block.
 * <p>
 * 
 * <tr valign="top">
 * <td>Phrases File
 * <td>Specifies a file containing phrases of text to be presented to participants
 * for entry. Phrases are drawn from the file at random. Typically, <a
 * href="phrases2.txt">phrases2.txt</a> is used. This is the phrase set published by MacKenzie and
 * Soukoreff in 2003 (see Related references, above). Other file names may be specified. For
 * example, the file <a href="quickbrownfox.txt">quickbrownfox.txt</a> contains the phrase
 * <p>
 * 
 * <pre>
*     the quick brown fox jumps over the lazy dog
* </pre>
 * 
 * which might be useful for demonstration or specialized testing.
 * <p>
 * 
 * <tr valign="top">
 * <td>Entry Mode
 * <td>The entry mode (see above).
 * <p>
 * 
 * <tr valign="top">
 * <td>Dictionary
 * <td>Used for the T9 entry modes, which require a built-in word-frequency dictionary,
 * appended with the letter-to-key mappings for the phone keypad. For this, two files are provided.
 * One is based on the BNC "d1" file with 9022 words. The file is modified slightly, adding words
 * from phrases2.txt that are not in the core dictionary file. The resulting file contains 9195
 * words: <a href="d1phrases2-wordfreq-phoneks.txt">d1phrases2-wordfreq-phoneks.txt</a>. Using this
 * version of the d1 dictionary guarantees that all words in phrases2.txt can be entered using the
 * T9 modes. A second dictionary file uses the expanded (d2) version of the BNC with 64,566 words.
 * There are 64,589 unique words when the excluded words from phrases2.txt are added. The file is <a
 * href="d2phrases2-wordfreq-phoneks.txt">d2phrases2-wordfreq-phoneks.txt</a>.
 * <p>
 * 
 * <tr valign="top">
 * <td>Size of Candidate List
 * <td>Used for the word completion modes, where a candidate list is
 * produced after each keystroke. The size of the candidate list is configurable as 0, 1, 2, 5, or
 * 10 words. The list is partitioned in two parts with words exactly matching the current key
 * sequence appearing first, followed by extended words, where the current key sequence is treated
 * as a word stem.
 * <p>
 * 
 * <tr valign="top">
 * <td>Show Candidates After
 * <td>Indicates the number of keystroke after which the candidate list is
 * shown to the participant
 * <p>
 * 
 * <tr valign="top">
 * <td>Out-of-dictionary Beep
 * <td>Used for the T9WC and T9WCWP modes. If enabled, the system will beep
 * if a keystroke results in no dictionary matches for the current keystroke sequence. NOTE: It is
 * not meaningful to implement this mode for the basic T9 mode, since it is possible for a correct
 * keystroke sequence to be "out of dictionary" for a portion of a correctly inputed word. For
 * example, depending on the dictionary, the word "computer" may have the following tentative
 * interpretation during entry:
 * <p>
 * 
 * <pre>
*     2 a
*     6 an
*     6 con
*     7 cons
*     8 boost
*     8 266788
*     3 compute
*     7 computer
* </pre>
 * 
 * After the key sequence 266788, there is no dictionary match. It would not be meaningful to
 * produce an out-of-dictionary beep in basic T9 mode, since the input is correct. In the T9WC or
 * T9WCWP modes, however, there would be a legitimate "tentative" match for 266788 ("comput") so
 * enabling the "out-of-dictionary beep" mode would not generate an erroneous beep.
 * <p>
 * </table>
 * </blockquote>
 * 
 * <h3>Embedded Parameters</h3>
 * <p>
 * 
 * The program includes some embedded parameters that are hard-coded into the source file:
 * 
 * <ul>
 * <li>Timeout - used to segment consecutive letters on the same key during multiptap entry. The
 * timeout is set to 1500 ms. NOTE: Segmenting can also be performed using a timeout kill keypress
 * (see below).
 * <p>
 * 
 * <li>Word prediction list - used for the word prediction mode. This is a list of the ten most
 * common words in English (the, of, and, a, in, to, is, was, it, for) which are presented prior to
 * the first keystroke of each word. These words constitute about 28% of English; so, even this
 * simple mode of word prediction offers a considerable reduction in KSPC.
 * </ul>
 * <p>
 * 
 * Note: The parameter settings for Dictionary, Size_of_candidate_list, and Show_candidates_after
 * are not included in the output data files. If any of these parameters is manipulated as a factor
 * in an experiment, the levels of the factor should be assigned codes using the Condition_code
 * parameter.
 * <p>
 * 
 * <h3>Program Operation</h3>
 * <p>
 * 
 * The program adopts a set of behaviours, according to the entry mode and setup parameters, then
 * launches with a GUI somewhat as follows:
 * <p>
 * 
 * <center><a href="PhoneKeypadExperiment-1.jpg"><img src="PhoneKeypadExperiment-1.jpg"
 * width="500"></a></center>
 * <p>
 * 
 * The keypad operates as expected, with letters entered using keys 2 through 9 and a SPACE entered
 * using 0. Additionally, three keys are used for the BACKSPACE, TIMEOUT KILL, and NEXT functions:
 * <p>
 * 
 * <center><img src="PhoneKeypadExperiment-3.jpg"></center>
 * <p>
 * 
 * The keypad rendering is typical of many mobile phones, so the interpretation of the above keys
 * may not be apparent to the user. To assist, tooltip help is implemented.
 * <p>
 * 
 * After a word is entered with the letter keys, clicking SPACE (0) commits the word and appends a
 * space in preparation for the next word.
 * <p>
 * 
 * In the word completion and prediction modes, a word may be selected early by clicking on it in
 * the candidate list.
 * <p>
 * 
 * When the BEGIN button is clicked, the first phrase is presented. Timing begins on the next key
 * selection. At the end of a phrase, the ENTER button is clicked; thus ending the timing for that
 * phrase. Participants may rest at their discretion between phrases.
 * <p>
 * 
 * Summary data appear in a popup dialog after entry of each phrase. Below is an example:
 * <p>
 * 
 * <center><a href="PhoneKeypadExperiment-2.jpg"><img src="PhoneKeypadExperiment-2.jpg"
 * width="400"></a></center>
 * <p>
 * 
 * <h3>Output Data Files</h3>
 * 
 * The data are also written to output summary data files. There are two output data files: the
 * summary data 1 file (.sd1) contains timestamps and keystroke-level information, and the summary
 * data 2 file (.sd2) contains aggregate data at the phrase-level, one line per phrase. Here is an
 * example of each:
 * <p>
 * 
 * <ul>
 * <li><a href="PhoneKeypadExperiment-sd1-example.txt">sd1 example<a/>
 * <li><a href="PhoneKeypadExperiment-sd2-example.txt">sd2 example<a/>
 * </ul>
 * <p>
 * 
 * The data in the sd2 file are full-precision, comma-delimited. Importing into a spreadsheet
 * application provides a convenient method to examine the data on a phrase-by-phrase basis. 
 * Below is an example of an sd2 file after importing into Microsoft <i>Excel</i>: (click to enlarge)
 * <p>
 * 
 * <center>
 * <a href="PhoneKeypadExperiment-4.jpg"><img src="PhoneKeypadExperiment-4.jpg" width=700></a>
 * </center>
 * <p>
 * 
 * Actual output files use "PhoneKeypadExperiment" as the base filename. This is followed by the
 * participant code, the condition code, the block code, and the entry mode, for example,
 * <code>PhoneKeypadExperiment-P99-C01-T9WCWP-S01-B01.sd2</code>.
 * <p>
 * 
 * <h3>Miscellaneous</h3>
 * 
 * When using this program in an experiment, it is a good idea to terminate all other applications
 * and disable the system's network connection. This will maintain the integrity of the data
 * collected and ensure that the program runs without hesitations.
 * <p>
 * 
 * @author Scott MacKenzie, 2009-2015
 * @author Steven Castellucci, 2014
 */
public class PhoneKeypadExperiment
{
	final static String APP_NAME = "PhoneKeypadExperiment";
	final static int PARAMETERS = 10; // must equal the number of parameters defined below

	// setup parameters (1st value is default)
	final static String[] PARTICIPANT_CODES = { "P00", "P01", "P02", "P03", "P04", "P05", "P06", "P07", "P08", "P09",
			"P10", "P11", "P12", "P13", "P14", "P15", "P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23", "P24",
			"P25", "P26", "P27", "P28", "P29", "P30" };
	final static String[] CONDITION_CODE = { "C00" };
	final static String[] SESSION_CODES = { "S00", "S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08", "S09",
			"S10", "S11", "S12", "S13", "S14", "S15", "S16", "S17", "S18", "S19", "S20", "S21", "S22", "S23", "S24",
			"S25", "S26", "S27", "S28", "S29", "S30" };
	final static String[] NUMBER_OF_PHRASES = { "5" };
	final static String[] PHRASES_FILE = { "phrases2.txt" };
	final static String[] ENTRY_MODE = { "MULTITAP", "T9", "T9WC", "T9WCWP" };
	final static String[] DICTIONARY = { "d2phrases2-wordfreq-phoneks.txt" };
	final static String[] SIZE_OF_CANDIDATE_LIST = { "0", "1", "2", "5", "10" };
	final static String[] SHOW_CANDIDATE_LIST_AFTER = { "1", "2", "3", "4", "5" };
	final static String[] OUT_OF_DICTIONARY_BEEP = { "true" };

	// identify other files needed (used when execution from the .jar file for the 1st time)
	final static String[] OTHER_FILES = { "phrases2.txt", "quickbrownfox.txt", "d1phrases2-wordfreq-phoneks.txt",
			"d2phrases2-wordfreq-phoneks.txt" };

	public static void main(String[] args) throws IOException
	{
		// use Win32 look and feel
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e)
		{
		}

		/*
		 * Define the default parameters settings. These are only used if the .cfg file does not
		 * exit, such as when the application is launched for the 1st time.
		 */
		String[] defaultParameters = new String[PARAMETERS];
		defaultParameters[0] = PARTICIPANT_CODES[0];
		defaultParameters[1] = CONDITION_CODE[0];
		defaultParameters[2] = SESSION_CODES[0];
		defaultParameters[3] = NUMBER_OF_PHRASES[0];
		defaultParameters[4] = PHRASES_FILE[0];
		defaultParameters[5] = ENTRY_MODE[0];
		defaultParameters[6] = DICTIONARY[0];
		defaultParameters[7] = SIZE_OF_CANDIDATE_LIST[0];
		defaultParameters[8] = SHOW_CANDIDATE_LIST_AFTER[0];
		defaultParameters[9] = OUT_OF_DICTIONARY_BEEP[0];

		Configuration c = Configuration.readConfigurationData(APP_NAME, defaultParameters, OTHER_FILES);
		if (c == null)
		{
			System.out.println("Error reading configuration data from " + APP_NAME + ".cfg!");
			System.exit(0);
		}

		// ----------------------------------------------------------
		// user the app's parameter settings to create a setup dialog
		// ----------------------------------------------------------

		SetupItemInfo[] sii = new SetupItemInfo[PARAMETERS];
		sii[0] = new SetupItemInfo(SetupItem.COMBO_BOX, "Participant code ", PARTICIPANT_CODES);
		sii[1] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Condition code ", CONDITION_CODE);
		sii[2] = new SetupItemInfo(SetupItem.COMBO_BOX, "Session code ", SESSION_CODES);
		sii[3] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Number of phrases ", NUMBER_OF_PHRASES);
		sii[4] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Phrases file ", PHRASES_FILE);
		sii[5] = new SetupItemInfo(SetupItem.COMBO_BOX, "Entry mode ", ENTRY_MODE);
		sii[6] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Dictionary ", DICTIONARY);
		sii[7] = new SetupItemInfo(SetupItem.COMBO_BOX, "Size of candidate list ", SIZE_OF_CANDIDATE_LIST);
		sii[8] = new SetupItemInfo(SetupItem.COMBO_BOX, "Show candidate after  ", SHOW_CANDIDATE_LIST_AFTER);
		sii[9] = new SetupItemInfo(SetupItem.CHECK_BOX, "Out-of-dictionary beep  ", OUT_OF_DICTIONARY_BEEP);

		// use setup to allow changes to the default or existing configuration
		Setup s = new Setup(null, c, APP_NAME, sii);
		s.showSetup(null);

		PhoneKeypadExperimentGui screen = new PhoneKeypadExperimentGui(c);
		screen.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		screen.setTitle("Phone Keypad Experiment");
		screen.pack();

		// put application in center of screen
		int w = screen.getWidth();
		int h = screen.getHeight();
		Toolkit t = Toolkit.getDefaultToolkit();
		Dimension d = t.getScreenSize();
		screen.setLocation((d.width - w) / 2, (d.height - h) / 2);
		screen.setVisible(true);
	}
}

class PhoneKeypadExperimentGui extends JFrame implements ActionListener, ItemListener, MouseListener
{
	private static final long serialVersionUID = 1L;

	final String SD1_HEADER = "";
	final String SD2_HEADER = "App,Participant,Condition,EntryMode,Session,Block,Phrase,Keystrokes,Characters,Time_(s),MSD,Speed_(wpm),Error_rate_(%),KSPC\n";

	// setup parameters
	String participantCode, conditionCode, sessionCode;
	String blockCode;
	int numberOfPhrases;
	String phrasesFile;
	String entryMode;
	String dictionaryFile;
	int sizeOfCandidateList;
	int showCandidateListAfter;
	boolean outOfDictionaryBeep;

	int phrasesRemaining;
	boolean phoneLayout = false;
	boolean qwertyLayout = false;
	Timer t;
	int tapCount = 0;
	String newKey = "";
	String lastKey = "";
	String m = "";

	String build = "";
	String[] candidates = null;

	Color enabledBackgroundColor = new Color(220, 220, 220);
	Color disabledBackgroundColor = new Color(180, 180, 180);
	int candidateIndex;

	final String LOWERCASE = "abc";
	final String UPPERCASE = "ABC";
	String caseMode = LOWERCASE;
	JLabel caseLabel = new JLabel();
	JLabel entryModeLabel = new JLabel();

	EncodedWord[] ew;
	final int MAX_CANDIDATES = 10;
	JTextField[] queries = new JTextField[MAX_CANDIDATES];
	TextMessage tm = new TextMessage();

	final Font TEXT_FIELD_FONT = new Font("sansserif", Font.BOLD, 18);
	String[] keyLetters = new String[10];
	String[] keyLabels = new String[12];

	String[] keyActionCommand = { "1", "2", "3", "4", "5", "6", "7", "8", "9", "*", "0", "#" };
	JButton[] key = new JButton[keyLabels.length];
	JButton enter = new JButton("Enter");
	JButton begin = new JButton("Begin");
	// 6 icon buttons at top of keypad (like a mobile phone)
	JButton[] iconButtons = new JButton[6];
	ImageIcon[] iconImages = { new ImageIcon("PhoneKeypadExperiment-leftselect.gif"),
			new ImageIcon("PhoneKeypadExperiment-up.gif"), new ImageIcon("PhoneKeypadExperiment-rightselect.gif"),
			new ImageIcon("PhoneKeypadExperiment-pickup.gif"), new ImageIcon("PhoneKeypadExperiment-down.gif"),
			new ImageIcon("PhoneKeypadExperiment-hangup.gif") };
	String[] iconActionCommand = { "leftselect", "up", "rightselect", "pickup", "down", "hangup" };

	JPanel keypadPanel;
	JPanel outerPanel;
	JPanel candidatePanel;

	JTextField presentedText;
	JTextField transcribedText;
	Random r = new Random();
	Vector<Sample> samples;
	// JTextArea results;
	long t1, t2;
	int count; // count keystrokes per phrase
	String[] phrases;
	BufferedWriter sd1File, sd2File;
	Configuration c;

	Clip outOfDictionaryClip;
	Clip ambiguousSelectionClip;

	JDialog resultsDialog;
	JOptionPane resultsPane;
	JTextArea resultsArea;

	// constructor
	PhoneKeypadExperimentGui(Configuration cArg) throws IOException
	{
		// Initialize setup parameters
		c = cArg;
		participantCode = c.getConfigurationParameter(0);
		conditionCode = c.getConfigurationParameter(1);
		sessionCode = c.getConfigurationParameter(2);
		// Note: block code computed automatically in software (see below)
		numberOfPhrases = Integer.parseInt(c.getConfigurationParameter(3));
		phrasesFile = c.getConfigurationParameter(4);
		entryMode = c.getConfigurationParameter(5);
		dictionaryFile = c.getConfigurationParameter(6);
		sizeOfCandidateList = Integer.parseInt(c.getConfigurationParameter(7));
		showCandidateListAfter = Integer.parseInt(c.getConfigurationParameter(8));
		outOfDictionaryBeep = Boolean.valueOf(c.getConfigurationParameter(9));

		// initialize sound clips
		outOfDictionaryClip = initSound("miss.wav");
		ambiguousSelectionClip = initSound("undo.wav");

		phrasesRemaining = numberOfPhrases;
		phoneLayout = true; // convert to setup arg later, perhaps
		enter.setFont(TEXT_FIELD_FONT);
		begin.setFont(TEXT_FIELD_FONT);

		caseLabel.setFont(new Font("sansserif", Font.BOLD, 14));
		entryModeLabel.setFont(new Font("sansserif", Font.BOLD, 14));
		caseLabel.setText(caseMode);
		entryModeLabel.setText(entryMode);

		String[] qwertyLetters = { "1qwer", "2tyui", "3op", "4as", "5df", "6ghjkl", "7zxcv", "8bn", "9m", "0 " };
		String[] phoneLetters = { "1.,?!-&'", "2abc", "3def", "4ghi", "5jkl", "6mno", "7pqrs", "8tuv", "9wxyz", "0 " };

		String[] qwertyLabels = { "1 qwer", "2 tyui", "3 op", "4 as", "5 df", "6 ghjkl", "7 zxcv", "8 bn", "9 m", "*",
				"0 _", "#" };

		String[] phoneLabels = { "1", "2 abc", "3 def", "4 ghi", "5 jkl", "6 mno", "7 pqrs", "8 tuv", "9 wxyz", "*",
				"0 _", "#" };

		for (int i = 0; i < 10; ++i)
			if (phoneLayout)
				keyLetters[i] = phoneLetters[i];
			else
				keyLetters[i] = qwertyLetters[i];

		for (int i = 0; i < 12; ++i)
			if (phoneLayout)
				keyLabels[i] = phoneLabels[i];
			else
				keyLabels[i] = qwertyLabels[i];

		samples = new Vector<Sample>();
		t = new Timer(1500, this); // start when Begin button clicked

		candidateIndex = 0;

		enter.addActionListener(this);
		begin.addActionListener(this);

		// load dictionary
		// String file = "d1phrases2-wordfreq-phoneks.txt"; // d1 with phrases2
		// words added
		String file = dictionaryFile; // from config file or setup dialog
		ew = EncodedWord.loadCodedDictionary(file);

		// open disk file for input
		BufferedReader inFile = new BufferedReader(new FileReader(phrasesFile));
		Vector<String> v = new Vector<String>();
		String s;
		while ((s = inFile.readLine()) != null)
			v.addElement(s);
		phrases = new String[v.size()];
		v.copyInto(phrases);

		// -----------------------
		// initialize output files
		// -----------------------

		blockCode = "";
		int blockCodeNumber = 0;
		String base = PhoneKeypadExperiment.APP_NAME + "-" + participantCode + "-" + conditionCode + "-" + entryMode
				+ "-" + sessionCode;

		String s1 = "";
		String s2 = "";
		do // find next available block code
		{
			++blockCodeNumber;
			blockCode = blockCodeNumber < 10 ? "B0" + blockCodeNumber : "B" + blockCodeNumber;
			s1 = base + "-" + blockCode + ".sd1";
			s2 = base + "-" + blockCode + ".sd2";
		} while (new File(s1).exists());

		try
		{
			sd1File = new BufferedWriter(new FileWriter(s1));
			sd2File = new BufferedWriter(new FileWriter(s2));
		} catch (IOException e)
		{
			System.out.println("I/O error: can't open sd1/sd2 data files");
			System.exit(0);
		}

		try
		{
			sd1File.write(SD1_HEADER, 0, SD1_HEADER.length());
			sd1File.flush();
			sd2File.write(SD2_HEADER, 0, SD2_HEADER.length());
			sd2File.flush();
		} catch (IOException e)
		{
			System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
			System.exit(1);
		}

		// initialize GUI components

		setBackground(Color.white);
		presentedText = new JTextField(10);
		transcribedText = new JTextField(10);

		presentedText.setFont(TEXT_FIELD_FONT);
		presentedText.setEditable(false);
		transcribedText.setFont(TEXT_FIELD_FONT);
		transcribedText.setEditable(true); // allows insertion tracker to appear
		presentedText.setBackground(Color.WHITE);
		transcribedText.setBackground(Color.WHITE);
		presentedText.setForeground(new Color(0, 0, 128));

		initCandidateList();
		initKeypad();

		// prepare results dialog
		resultsArea = new JTextArea(15, 40);
		resultsArea.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsArea.setBackground((new JButton()).getBackground());
		resultsPane = new JOptionPane(resultsArea, JOptionPane.INFORMATION_MESSAGE);
		resultsPane.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsDialog = resultsPane.createDialog(this, "Information");

		JPanel p1 = new JPanel();
		p1.add(outerPanel);
		p1.add(candidatePanel);
		p1.setBorder(BorderFactory.createEmptyBorder(25, 25, 25, 25));

		JPanel p4 = new JPanel();
		p4.setLayout(new FlowLayout(FlowLayout.LEFT, 10, 10));
		p4.add(entryModeLabel);
		p4.add(caseLabel);

		JPanel p2 = new JPanel();
		p2.setLayout(new BoxLayout(p2, BoxLayout.Y_AXIS));
		p2.add(Box.createRigidArea(new Dimension(0, 10)));
		p2.add(presentedText);
		p2.add(Box.createRigidArea(new Dimension(0, 10)));
		p2.add(transcribedText);
		p2.add(Box.createRigidArea(new Dimension(0, 10)));
		p2.add(p4);
		p2.setBorder(BorderFactory.createEmptyBorder(25, 25, 25, 25));

		JPanel p3 = new JPanel(new BorderLayout());
		p3.add(p2, "North");
		p3.add(p1, "Center");

		this.setContentPane(p3);
	}

	private void initKeypad()
	{
		keypadPanel = new JPanel();
		keypadPanel.setLayout(new GridLayout(6, 3));

		for (int i = 0; i < iconButtons.length; ++i)
		{
			iconButtons[i] = new JButton(iconImages[i]);
			iconButtons[i].setCursor(new Cursor(Cursor.HAND_CURSOR));
			iconButtons[i].setActionCommand(iconActionCommand[i]);
			iconButtons[i].addActionListener(this);
			keypadPanel.add(iconButtons[i]);
		}
		iconButtons[2].setToolTipText("Backspace key");
		iconButtons[4].setToolTipText("Timeout kill key");

		for (int i = 0; i < keyLabels.length; ++i)
		{
			key[i] = new JButton(keyLabels[i]);
			key[i].setCursor(new Cursor(Cursor.HAND_CURSOR));
			key[i].setFont(new Font("SansSerif", Font.BOLD, 16));
			key[i].setMargin(new Insets(3, 3, 3, 3));
			key[i].setActionCommand(keyActionCommand[i]);
			key[i].addActionListener(this);
			keypadPanel.add(key[i]);
		}
		key[9].setToolTipText("Next key");

		JPanel beginPanel = new JPanel(new FlowLayout(FlowLayout.CENTER));
		beginPanel.add(begin);
		JPanel enterPanel = new JPanel(new FlowLayout(FlowLayout.CENTER));
		enterPanel.add(enter);

		outerPanel = new JPanel();
		outerPanel.setLayout(new BorderLayout(10, 10));
		outerPanel.add(beginPanel, "West");
		outerPanel.add(keypadPanel, "Center");
		outerPanel.add(enterPanel, "East");

		Border b1 = new EtchedBorder();
		Border b2 = BorderFactory.createEmptyBorder(5, 5, 5, 5);
		outerPanel.setBorder(new CompoundBorder(b1, b2));
	}

	private void initCandidateList()
	{
		candidatePanel = new JPanel(new GridLayout(MAX_CANDIDATES, 1));
		for (int i = 0; i < MAX_CANDIDATES; ++i)
		{
			queries[i] = new JTextField(10);
			queries[i].setEditable(false);
			queries[i].setCursor(new Cursor(Cursor.HAND_CURSOR));
			queries[i].setFont(TEXT_FIELD_FONT);
			candidatePanel.add(queries[i]);
			if (entryMode.equals("T9WC") || entryMode.equals("T9WCWP"))
			{
				if (i < sizeOfCandidateList)
				{
					queries[i].setBackground(enabledBackgroundColor);
					queries[i].addMouseListener(this);
				} else
				{
					queries[i].setBackground(disabledBackgroundColor);
					queries[i].removeMouseListener(this); // don't need this
				}
			}
		}
	}

	// *****************************
	// implement ItemListener method
	// *****************************

	@Override
	public void itemStateChanged(ItemEvent ie)
	{
	}

	// *******************************
	// implement ActionListener method
	// *******************************

	@Override
	public void actionPerformed(ActionEvent ae)
	{
		Object source = ae.getSource();

		// get text typed so far
		m = transcribedText.getText();

		// BEGIN button
		if (source == begin)
		{
			if (entryMode.equals("MULTITAP"))
				t.start();
			if (entryMode.equals("T9WCWP"))
				initQueries();
			presentedText.setText(phrases[r.nextInt(phrases.length)]);
			transcribedText.setText("");
			transcribedText.requestFocus();
			tm.clear();
			t1 = 0;
			count = 0;
			samples = new Vector<Sample>();
			return;
		}

		// ENTER button
		if (source == enter)
		{
			String s1 = presentedText.getText().toLowerCase().trim();
			String s2 = transcribedText.getText().toLowerCase().trim();

			MSD s1s2 = new MSD(s1, s2);
			int msd = s1s2.getMSD();

			// build formated results for GUI
			String results = "Thank you!\n\n";
			results += String.format("Presented:\t%s\n", s1);
			results += String.format("Transcribed:\t%s\n", s2);
			results += String.format("-----\n");
			results += String.format("Keystrokes:\t%d\n", count);
			results += String.format("Characters:\t%d\n", s2.length());
			results += String.format("Words:\t%.1f\n", s2.length() / 5.0);
			results += String.format("Time:\t%.2f s\n", t2 / 1000.0);
			results += String.format("MSD:\t%d\n", msd);
			results += String.format("-----\n");
			results += String.format("Speed:\t%.2f wpm\n", wpm(s2, t2));
			results += String.format("Error rate:\t%.1f%%\n", s1s2.getErrorRateNew());
			results += String.format("KSPC:\t%.4f\n", (double)count / s2.length());
			resultsArea.setText(results);
			resultsDialog.setVisible(true);

			// build output data for sd2 file

			String sd2Stuff = "PhoneKeypadExperiment," + participantCode + "," + conditionCode + "," + entryMode + ","
					+ sessionCode + "," + blockCode + "," + (numberOfPhrases - phrasesRemaining + 1) + ",";
			sd2Stuff += count + ","; // number of keystrokes
			sd2Stuff += s2.length() + ","; // number of characters entered
			sd2Stuff += t2 / 1000.0 + ","; // time in seconds
			sd2Stuff += msd + ","; // minimum string distance
			sd2Stuff += wpm(s2, t2) + ","; // speed in words per minute
			sd2Stuff += s1s2.getErrorRateNew() + ","; // error rate (based on MSD & alignments)
			sd2Stuff += (double)count / s2.length() + "\n"; // KSPC (end of line too!)

			// build output data for sd1 file
			String sd1Stuff = s1 + "\n";
			sd1Stuff += s2 + "\n";
			for (int i = 0; i < samples.size(); ++i)
				sd1Stuff += samples.elementAt(i) + "\n";
			sd1Stuff += "#\n";

			// prepare for next phrase
			samples = new Vector<Sample>();
			presentedText.setText(phrases[r.nextInt(phrases.length)]);
			tm.clear();
			clearQueries();
			build = "";
			transcribedText.setText("");
			transcribedText.requestFocus();
			t1 = 0;
			count = 0;
			if (entryMode.equals("T9WCWP"))
				initQueries();

			// write to data files
			try
			{
				sd2File.write(sd2Stuff, 0, sd2Stuff.length());
				sd2File.flush();
				sd1File.write(sd1Stuff, 0, sd1Stuff.length());
				sd1File.flush();
			} catch (IOException e)
			{
				showError("ERROR WRITING TO DATA FILE!\n" + e);
				System.exit(1);
			}

			// check if last phrase in block
			if (--phrasesRemaining == 0)
			{
				try
				{
					sd1File.close();
					sd2File.close();
				} catch (IOException e)
				{
					showError("ERROR CLOSING DATA FILE!\n" + e);
					System.exit(1);
				}
				System.exit(0);
			}
			return;
		}

		// If we get here, the 'event' is from a text entry button

		// ****************************************************************
		// T9 entry mode (perhaps with word completion and word prediction)
		// ****************************************************************

		newKey = ae.getActionCommand();
		if (entryMode.equals("T9") || entryMode.equals("T9WC") || entryMode.equals("T9WCWP"))
		{
			// first process icon buttons (6)
			for (int i = 0; i < iconButtons.length; ++i)
			{
				if (ae.getSource() == iconButtons[i])
				{
					if (i == 2) // 'rightselect' (delete)
					{
						if (build.length() > 0)
						{
							build = build.substring(0, build.length() - 1);
							updateQueries();
							if (candidates != null)
							{
								String tentative = tm.getMessage();
								if (tentative.length() != 0)
									tentative += " ";
								tentative += candidates[0].substring(0, build.length());
								// tentative += candidates[0];
								transcribedText.setText(tentative);
								transcribedText.requestFocus();
							}
						}
						if (build.length() == 0)
							if (entryMode.equals("T9WCWP"))
								initQueries();
							else
								clearQueries();
					}
					gatherMeasurements();
					return;
				}
			} // ...end process icon buttons

			// process keypad keys (12)
			switch (newKey.charAt(0))
			{
			/*** SPACE ***/
				case '0':
					if (candidates != null)
					{
						// play a sound if the word selected in an ambiguous word
						if (candidates.length > 1)
						{
							// This feature needs a bit more thought.
							// Perhaps it should only be used in basic T9 mode,
							// since there are ambiguous word stems in the
							// T9WC and T9WCWP modes
							// playSound(ambiguousSelectionClip);
						}
						tm.addWord(candidates[candidateIndex]);
						build = "";
						candidateIndex = 0;
						if (entryMode.equals("T9WCWP"))
							initQueries();
						else
							clearQueries();
						transcribedText.setText(tm.getMessage() + " ");
						transcribedText.requestFocus();
					}
					gatherMeasurements();
					break;

				/*** no implementation ***/
				case '#':
					gatherMeasurements();
					break;

				case '*': // pick next word in ambiguous list
					if (candidates != null)
					{
						++candidateIndex;
						if (candidateIndex >= candidates.length)
							candidateIndex = 0;
						build = candidates[candidateIndex];
						if (tm.getMessage().length() == 0)
						{
							transcribedText.setText(build);
							transcribedText.requestFocus();
						} else
						{
							transcribedText.setText(tm.getMessage() + " " + build);
							transcribedText.requestFocus();
						}
					}
					gatherMeasurements();
					break;

				/*** alpha keys ***/
				case '1': // ?
				case '2': // abc
				case '3': // def
				case '4': // ghi
				case '5': // jkl
				case '6': // mno
				case '7': // pqrs
				case '8': // tuv
				case '9': // wxyz
					build += newKey;
					updateQueries();
					if (candidates != null)
					{
						String tentative = tm.getMessage();
						if (tentative.length() != 0)
							tentative += " ";
						tentative += candidates[0].substring(0, build.length());
						transcribedText.setText(tentative);
					}
					transcribedText.requestFocus();
					gatherMeasurements();
			}
		}

		// *******************
		// Multitap entry mode
		// *******************

		else
		// must be a multitap
		{
			if (ae.getSource() == t)
			{
				tapCount = 0;
				lastKey = "timeout";
				return;
			}

			for (int i = 0; i < iconButtons.length; ++i)
			{
				if (ae.getSource() == iconButtons[i])
				{
					if (i == 2) // 'rightselect' (delete)
					{
						if (m.length() > 0)
						{
							transcribedText.setText(m.substring(0, m.length() - 1));
							transcribedText.requestFocus();
						}
						tapCount = 0;
						t.restart();
					} else if (i == 4) // timeout kill button
					{
						tapCount = 0;
						lastKey = "timeout kill";
						t.restart();
					}
					gatherMeasurements();
					return;
				}
			}

			newKey = ((JButton)ae.getSource()).getText(); // reassign to key label
			if (!lastKey.equals(newKey))
				tapCount = 0;
			++tapCount;
			switch (newKey.charAt(0))
			{
				case '#': // (change case)
					if (caseMode.equals(LOWERCASE))
					{
						caseMode = UPPERCASE;
						caseLabel.setText(UPPERCASE);
					} else
					{
						caseMode = LOWERCASE;
						caseLabel.setText(LOWERCASE);
					}
					gatherMeasurements();
					break;

				case '1': // .,?!-&'1
				case '2': // abc2
				case '3': // def3
				case '4': // ghi4
				case '5': // jkl5
				case '6': // mno6
				case '7': // pqrs7
				case '8': // tuv8
				case '9': // wxyz9
				case '*': // *
				case '0': // _0
					for (int i = 0; i < keyLetters.length; ++i)
					{
						if (newKey.charAt(0) == keyLetters[i].charAt(0))
						{
							if (tapCount >= keyLetters[i].length())
								tapCount = 0;
							char tmp = keyLetters[i].charAt(tapCount);
							if (caseMode == UPPERCASE)
								tmp = Character.toUpperCase(tmp);
							if (newKey == lastKey)
								m = m.substring(0, m.length() - 1);
							transcribedText.setText(m + tmp);
						}
					}
					transcribedText.requestFocus();
					gatherMeasurements();
			}
			lastKey = newKey;
			t.restart();
		}
	} // end of actionPerformed

	private void showError(String msg)
	{
		JOptionPane.showMessageDialog(null, msg, "I/O Error", JOptionPane.ERROR_MESSAGE);
	}

	public void gatherMeasurements()
	{
		++count;
		if (t1 == 0)
			t1 = System.currentTimeMillis();
		t2 = System.currentTimeMillis() - t1;
		samples.addElement(new Sample(t2, newKey.toLowerCase()));
	}

	// *******************************
	// implement MouseListener methods
	// *******************************

	public void mouseClicked(MouseEvent me)
	{
		Object source = me.getSource();
		int i = 0;
		for (i = 0; i < MAX_CANDIDATES; ++i)
			if (source == queries[i])
				break;

		// if the user clicks on an empty candidate field, exit now (to avoid exception)
		if (candidates == null || i > candidates.length - 1)
			return;

		if (sizeOfCandidateList == 1)
			tm.addWord(candidates[candidateIndex]);
		else
			tm.addWord(candidates[i]);
		build = "";
		if (entryMode.equals("T9WCWP"))
			initQueries();
		else
			clearQueries();
		transcribedText.setText(tm.getMessage() + " ");
		transcribedText.requestFocus();
		newKey = "Candidate " + i;
		gatherMeasurements();
	}

	public void mouseEntered(MouseEvent me)
	{
	}

	public void mouseExited(MouseEvent me)
	{
	}

	public void mousePressed(MouseEvent me)
	{
	}

	public void mouseReleased(MouseEvent me)
	{
	}

	// update queries, as per 'build'
	private void updateQueries()
	{
		String[] tmp = null;
		if (entryMode.equals("T9WC") || entryMode.equals("T9WCWP"))
		{
			tmp = EncodedWord.getQuery(build, ew);
		} else
		// 'plain' T9 (i.e., no word completion)
		{
			tmp = EncodedWord.getUnique(build, ew);
			if (tmp == null)
			{
				tmp = new String[1];
				tmp[0] = build;
			}
		}

		if (tmp == null && build.length() > 0)
		{
			build = build.substring(0, build.length() - 1);
			if (outOfDictionaryBeep)
			{
				playSound(outOfDictionaryClip);
			}
		} else if (tmp != null)
		{
			candidates = tmp;
			for (int i = 0; i < sizeOfCandidateList; ++i)
				if (i < candidates.length && build.length() >= showCandidateListAfter
						&& (entryMode.equals("T9WC") || entryMode.equals("T9WCWP")))
					queries[i].setText(candidates[i]);
				else
					queries[i].setText("");
		}
	}

	private void clearQueries()
	{
		for (int i = 0; i < MAX_CANDIDATES; ++i)
			queries[i].setText("");
		candidates = null;
	}

	private void initQueries()
	{
		String[] s = EncodedWord.getMostFrequent(sizeOfCandidateList, ew);
		candidates = new String[s.length];
		for (int i = 0; i < s.length; ++i)
		{
			queries[i].setText(s[i]);
			candidates[i] = s[i];
		}
	}

	// compute typing speed in wpm given text entered and time in ms
	public static double wpm(String text, long msTime)
	{
		double speed = text.length();
		speed = speed / (msTime / 1000.0) * (60 / 5);
		return speed;
	}

	private class Sample
	{
		private long time;
		private String key;

		Sample(long timeArg, String keyArg)
		{
			time = timeArg;
			key = keyArg;
		}

		@Override
		public String toString()
		{
			return time + ", " + key;
		}
	}

	// Initialize stream for sound clip
	public Clip initSound(String soundFile)
	{
		AudioInputStream audioInputStream;
		Clip c = null;
		try
		{
			// Added for executable Jar:
			audioInputStream = AudioSystem.getAudioInputStream(new BufferedInputStream(getClass().getResourceAsStream(
					soundFile)));
			c = AudioSystem.getClip();
			c.open(audioInputStream);
		} catch (Exception e)
		{
			showError("ERROR: Unable to load sound clip (" + soundFile + ")");
		}
		return c;
	}

	// play sound
	public void playSound(Clip c)
	{
		if (c != null)
		{
			c.setFramePosition(0); // rewind clip to beginning
			c.start(); // stops at end of clip
		}
	}
}